@@ -35,6 +35,10 @@
 #include <cmath>
 
 #include "tscore/ink_config.h"
+#include <openssl/pem.h>
+
+#include "InkAPIInternal.h" // Added to include the ssl_hook and lifestyle_hook definitions
+
 #include "tscore/ink_platform.h"
 #include "tscore/I_Layout.h"
 #include "records/I_RecHttp.h"
@@ -43,14 +47,17 @@
 
 #include "P_Net.h"
 #include "P_SSLClientUtils.h"
+#include "P_SSLSNI.h"
 #include "P_SSLCertLookup.h"
 #include "P_SSLSNI.h"
+#include "P_TLSKeyLogger.h"
 #include "SSLDiags.h"
 #include "SSLSessionCache.h"
 #include "SSLSessionTicket.h"
 #include "YamlSNIConfig.h"
 
-int SSLConfig::configid                                     = 0;
+int SSLConfig::config_index                                 = 0;
+int SSLConfig::configids[]                                  = {0, 0};
 int SSLCertificateConfig::configid                          = 0;
 int SSLTicketKeyConfig::configid                            = 0;
 int SSLConfigParams::ssl_maxrecord                          = 0;
@@ -60,7 +67,10 @@ bool SSLConfigParams::ssl_ocsp_enabled                      = false;
 int SSLConfigParams::ssl_ocsp_cache_timeout                 = 3600;
 int SSLConfigParams::ssl_ocsp_request_timeout               = 10;
 int SSLConfigParams::ssl_ocsp_update_period                 = 60;
+char *SSLConfigParams::ssl_ocsp_user_agent                  = nullptr;
 int SSLConfigParams::ssl_handshake_timeout_in               = 0;
+int SSLConfigParams::origin_session_cache                   = 1;
+size_t SSLConfigParams::origin_session_cache_size           = 10240;
 size_t SSLConfigParams::session_cache_number_buckets        = 1024;
 bool SSLConfigParams::session_cache_skip_on_lock_contention = false;
 size_t SSLConfigParams::session_cache_max_bucket_size       = 100;
@@ -76,7 +86,6 @@ bool SSLConfigParams::server_allow_early_data_params = false;
 int SSLConfigParams::async_handshake_enabled = 0;
 char *SSLConfigParams::engine_conf_file      = nullptr;
 
-static std::unique_ptr<ConfigUpdateHandler<SSLCertificateConfig>> sslCertUpdate;
 static std::unique_ptr<ConfigUpdateHandler<SSLTicketKeyConfig>> sslTicketKey;
 
 SSLConfigParams::SSLConfigParams()
@@ -103,10 +112,12 @@ SSLConfigParams::reset()
   serverCertPathOnly = serverCertChainFilename = configFilePath = serverCACertFilename = serverCACertPath = clientCertPath =
     clientKeyPath = clientCACertFilename = clientCACertPath = cipherSuite = client_cipherSuite = dhparamsFile = serverKeyPathOnly =
       clientKeyPathOnly = clientCertPathOnly = nullptr;
+  ssl_ocsp_response_path_only                = nullptr;
   server_tls13_cipher_suites                 = nullptr;
   client_tls13_cipher_suites                 = nullptr;
   server_groups_list                         = nullptr;
   client_groups_list                         = nullptr;
+  keylog_file                                = nullptr;
   client_ctx                                 = nullptr;
   clientCertLevel = client_verify_depth = verify_depth = 0;
   verifyServerPolicy                                   = YamlSNIConfig::Policy::DISABLED;
@@ -141,10 +152,13 @@ SSLConfigParams::cleanup()
   client_cipherSuite      = static_cast<char *>(ats_free_null(client_cipherSuite));
   dhparamsFile            = static_cast<char *>(ats_free_null(dhparamsFile));
 
+  ssl_ocsp_response_path_only = static_cast<char *>(ats_free_null(ssl_ocsp_response_path_only));
+
   server_tls13_cipher_suites = static_cast<char *>(ats_free_null(server_tls13_cipher_suites));
   client_tls13_cipher_suites = static_cast<char *>(ats_free_null(client_tls13_cipher_suites));
   server_groups_list         = static_cast<char *>(ats_free_null(server_groups_list));
   client_groups_list         = static_cast<char *>(ats_free_null(client_groups_list));
+  keylog_file                = static_cast<char *>(ats_free_null(keylog_file));
 
   cleanupCTXTable();
   reset();
@@ -176,6 +190,67 @@ set_paths_helper(const char *path, const char *filename, char **final_path, char
   }
 }
 
+int
+UpdateServerPolicy(const char * /* name ATS_UNUSED */, RecDataT /* data_type ATS_UNUSED */, RecData data, void *cookie)
+{
+  SSLConfigParams *params = SSLConfig::acquire();
+  char *verify_server     = data.rec_string;
+  if (params != nullptr && verify_server != nullptr) {
+    Debug("ssl_load", "New Server Policy %s", verify_server);
+    params->SetServerPolicy(verify_server);
+  } else {
+    Debug("ssl_load", "Failed to load new Server Policy %p %p", verify_server, params);
+  }
+  return 0;
+}
+
+int
+UpdateServerPolicyProperties(const char * /* name ATS_UNUSED */, RecDataT /* data_type ATS_UNUSED */, RecData data, void *cookie)
+{
+  SSLConfigParams *params = SSLConfig::acquire();
+  char *verify_server     = data.rec_string;
+  if (params != nullptr && verify_server != nullptr) {
+    params->SetServerPolicyProperties(verify_server);
+  }
+  return 0;
+}
+
+void
+SSLConfigParams::SetServerPolicyProperties(const char *verify_server)
+{
+  if (strcmp(verify_server, "SIGNATURE") == 0) {
+    verifyServerProperties = YamlSNIConfig::Property::SIGNATURE_MASK;
+  } else if (strcmp(verify_server, "NAME") == 0) {
+    verifyServerProperties = YamlSNIConfig::Property::NAME_MASK;
+  } else if (strcmp(verify_server, "ALL") == 0) {
+    verifyServerProperties = YamlSNIConfig::Property::ALL_MASK;
+  } else if (strcmp(verify_server, "NONE") == 0) {
+    verifyServerProperties = YamlSNIConfig::Property::NONE;
+  } else {
+    Warning("%s is invalid for proxy.config.ssl.client.verify.server.properties.  Should be one of ALL, SIGNATURE, NAME, or NONE. "
+            "Default is ALL",
+            verify_server);
+    verifyServerProperties = YamlSNIConfig::Property::NONE;
+  }
+}
+
+void
+SSLConfigParams::SetServerPolicy(const char *verify_server)
+{
+  if (strcmp(verify_server, "DISABLED") == 0) {
+    verifyServerPolicy = YamlSNIConfig::Policy::DISABLED;
+  } else if (strcmp(verify_server, "PERMISSIVE") == 0) {
+    verifyServerPolicy = YamlSNIConfig::Policy::PERMISSIVE;
+  } else if (strcmp(verify_server, "ENFORCED") == 0) {
+    verifyServerPolicy = YamlSNIConfig::Policy::ENFORCED;
+  } else {
+    Warning("%s is invalid for proxy.config.ssl.client.verify.server.policy.  Should be one of DISABLED, PERMISSIVE, or ENFORCED. "
+            "Default is DISABLED",
+            verify_server);
+    verifyServerPolicy = YamlSNIConfig::Policy::DISABLED;
+  }
+}
+
 void
 SSLConfigParams::initialize()
 {
@@ -317,6 +392,8 @@ SSLConfigParams::initialize()
   ats_free(CACertRelativePath);
 
   // SSL session cache configurations
+  REC_ReadConfigInteger(ssl_origin_session_cache, "proxy.config.ssl.origin_session_cache");
+  REC_ReadConfigInteger(ssl_origin_session_cache_size, "proxy.config.ssl.origin_session_cache.size");
   REC_ReadConfigInteger(ssl_session_cache, "proxy.config.ssl.session_cache");
   REC_ReadConfigInteger(ssl_session_cache_size, "proxy.config.ssl.session_cache.size");
   REC_ReadConfigInteger(ssl_session_cache_num_buckets, "proxy.config.ssl.session_cache.num_buckets");
@@ -324,6 +401,8 @@ SSLConfigParams::initialize()
   REC_ReadConfigInteger(ssl_session_cache_timeout, "proxy.config.ssl.session_cache.timeout");
   REC_ReadConfigInteger(ssl_session_cache_auto_clear, "proxy.config.ssl.session_cache.auto_clear");
 
+  SSLConfigParams::origin_session_cache      = ssl_origin_session_cache;
+  SSLConfigParams::origin_session_cache_size = ssl_origin_session_cache_size;
   SSLConfigParams::session_cache_max_bucket_size =
     static_cast<size_t>(ceil(static_cast<double>(ssl_session_cache_size) / ssl_session_cache_num_buckets));
   SSLConfigParams::session_cache_skip_on_lock_contention = ssl_session_cache_skip_on_contention;
@@ -333,6 +412,10 @@ SSLConfigParams::initialize()
     session_cache = new SSLSessionCache();
   }
 
+  if (ssl_origin_session_cache == 1 && ssl_origin_session_cache_size > 0) {
+    origin_sess_cache = new SSLOriginSessionCache();
+  }
+
   // SSL record size
   REC_EstablishStaticConfigInt32(ssl_maxrecord, "proxy.config.ssl.max_record_size");
 
@@ -344,6 +427,7 @@ SSLConfigParams::initialize()
   REC_ReadConfigStringAlloc(ssl_ocsp_response_path, "proxy.config.ssl.ocsp.response.path");
   set_paths_helper(ssl_ocsp_response_path, nullptr, &ssl_ocsp_response_path_only, nullptr);
   ats_free(ssl_ocsp_response_path);
+  REC_ReadConfigStringAlloc(ssl_ocsp_user_agent, "proxy.config.http.request_via_str");
 
   REC_ReadConfigInt32(ssl_handshake_timeout_in, "proxy.config.ssl.handshake_timeout_in");
 
@@ -357,34 +441,14 @@ SSLConfigParams::initialize()
 
   char *verify_server = nullptr;
   REC_ReadConfigStringAlloc(verify_server, "proxy.config.ssl.client.verify.server.policy");
-  if (strcmp(verify_server, "DISABLED") == 0) {
-    verifyServerPolicy = YamlSNIConfig::Policy::DISABLED;
-  } else if (strcmp(verify_server, "PERMISSIVE") == 0) {
-    verifyServerPolicy = YamlSNIConfig::Policy::PERMISSIVE;
-  } else if (strcmp(verify_server, "ENFORCED") == 0) {
-    verifyServerPolicy = YamlSNIConfig::Policy::ENFORCED;
-  } else {
-    Warning("%s is invalid for proxy.config.ssl.client.verify.server.policy.  Should be one of DISABLED, PERMISSIVE, or ENFORCED",
-            verify_server);
-    verifyServerPolicy = YamlSNIConfig::Policy::DISABLED;
-  }
+  this->SetServerPolicy(verify_server);
   ats_free(verify_server);
+  REC_RegisterConfigUpdateFunc("proxy.config.ssl.client.verify.server.policy", UpdateServerPolicy, nullptr);
 
   REC_ReadConfigStringAlloc(verify_server, "proxy.config.ssl.client.verify.server.properties");
-  if (strcmp(verify_server, "SIGNATURE") == 0) {
-    verifyServerProperties = YamlSNIConfig::Property::SIGNATURE_MASK;
-  } else if (strcmp(verify_server, "NAME") == 0) {
-    verifyServerProperties = YamlSNIConfig::Property::NAME_MASK;
-  } else if (strcmp(verify_server, "ALL") == 0) {
-    verifyServerProperties = YamlSNIConfig::Property::ALL_MASK;
-  } else if (strcmp(verify_server, "NONE") == 0) {
-    verifyServerProperties = YamlSNIConfig::Property::NONE;
-  } else {
-    Warning("%s is invalid for proxy.config.ssl.client.verify.server.properties.  Should be one of SIGNATURE, NAME, or ALL",
-            verify_server);
-    verifyServerProperties = YamlSNIConfig::Property::NONE;
-  }
+  this->SetServerPolicyProperties(verify_server);
   ats_free(verify_server);
+  REC_RegisterConfigUpdateFunc("proxy.config.ssl.client.verify.server.properties", UpdateServerPolicyProperties, nullptr);
 
   ssl_client_cert_filename = nullptr;
   ssl_client_cert_path     = nullptr;
@@ -408,6 +472,13 @@ SSLConfigParams::initialize()
 
   REC_ReadConfigStringAlloc(client_groups_list, "proxy.config.ssl.client.groups_list");
 
+  REC_ReadConfigStringAlloc(keylog_file, "proxy.config.ssl.keylog_file");
+  if (keylog_file == nullptr) {
+    TLSKeyLogger::disable_keylogging();
+  } else {
+    TLSKeyLogger::enable_keylogging(keylog_file);
+  }
+
   REC_ReadConfigInt32(ssl_allow_client_renegotiation, "proxy.config.ssl.allow_client_renegotiation");
 
   REC_ReadConfigInt32(ssl_misc_max_iobuffer_size_index, "proxy.config.ssl.misc.io.max_buffer_index");
@@ -427,6 +498,30 @@ SSLConfigParams::getClientSSL_CTX() const
   return client_ctx;
 }
 
+int
+SSLConfig::get_config_index()
+{
+  return config_index;
+}
+
+int
+SSLConfig::get_loading_config_index()
+{
+  return config_index == 0 ? 1 : 0;
+}
+
+void
+SSLConfig::commit_config_id()
+{
+  // Update the active config index
+  config_index = get_loading_config_index();
+
+  if (configids[get_loading_config_index()] != 0) {
+    // Start draining to free the old config
+    configProcessor.set(configids[get_loading_config_index()], nullptr);
+  }
+}
+
 void
 SSLConfig::startup()
 {
@@ -436,34 +531,44 @@ SSLConfig::startup()
 void
 SSLConfig::reconfigure()
 {
-  Debug("ssl", "Reload SSLConfig");
+  Debug("ssl_load", "Reload SSLConfig");
   SSLConfigParams *params;
   params = new SSLConfigParams;
+  // start loading the next config
+  int loading_config_index        = get_loading_config_index();
+  configids[loading_config_index] = configProcessor.set(configids[loading_config_index], params);
   params->initialize(); // re-read configuration
-  configid = configProcessor.set(configid, params);
+  // Make the new config available for use.
+  commit_config_id();
 }
 
 SSLConfigParams *
 SSLConfig::acquire()
 {
-  return static_cast<SSLConfigParams *>(configProcessor.get(configid));
+  return static_cast<SSLConfigParams *>(configProcessor.get(configids[get_config_index()]));
+}
+
+SSLConfigParams *
+SSLConfig::load_acquire()
+{
+  return static_cast<SSLConfigParams *>(configProcessor.get(configids[get_loading_config_index()]));
 }
 
 void
 SSLConfig::release(SSLConfigParams *params)
 {
-  configProcessor.release(configid, params);
+  configProcessor.release(configids[get_config_index()], params);
+}
+
+void
+SSLConfig::load_release(SSLConfigParams *params)
+{
+  configProcessor.release(configids[get_loading_config_index()], params);
 }
 
 bool
 SSLCertificateConfig::startup()
 {
-  sslCertUpdate.reset(new ConfigUpdateHandler<SSLCertificateConfig>());
-  sslCertUpdate->attach("proxy.config.ssl.server.multicert.filename");
-  sslCertUpdate->attach("proxy.config.ssl.server.cert.path");
-  sslCertUpdate->attach("proxy.config.ssl.server.private_key.path");
-  sslCertUpdate->attach("proxy.config.ssl.server.cert_chain.filename");
-  sslCertUpdate->attach("proxy.config.ssl.server.session_ticket.enable");
   // Exit if there are problems on the certificate loading and the
   // proxy.config.ssl.server.multicert.exit_on_load_fail is true
   SSLConfig::scoped_config params;
@@ -485,19 +590,22 @@ SSLCertificateConfig::reconfigure()
   // twice the healthcheck period to simulate a loading a large certificate set.
   if (is_action_tag_set("test.multicert.delay")) {
     const int secs = 60;
-    Debug("ssl", "delaying certificate reload by %d secs", secs);
+    Debug("ssl_load", "delaying certificate reload by %d secs", secs);
     ink_hrtime_sleep(HRTIME_SECONDS(secs));
   }
 
   SSLMultiCertConfigLoader loader(params);
-  loader.load(lookup);
+  if (!loader.load(lookup)) {
+    retStatus = false;
+  }
 
   if (!lookup->is_valid) {
     retStatus = false;
   }
-  // If there are errors in the certificate configs and we had wanted to exit on error
-  // we won't want to reset the config
-  if (retStatus || !params->configExitOnLoadError) {
+
+  // If there are errors in the certificate configs force the load anyway if there
+  // is no configuration at all (i.e. this is the initial load).
+  if (retStatus || 0 == configid) {
     configid = configProcessor.set(configid, lookup);
   } else {
     delete lookup;
@@ -521,6 +629,9 @@ SSLCertificateConfig::acquire()
 void
 SSLCertificateConfig::release(SSLCertLookup *lookup)
 {
+  if (lookup == nullptr) {
+    return;
+  }
   configProcessor.release(configid, lookup);
 }
 
@@ -555,7 +666,7 @@ SSLTicketParams::LoadTicket(bool &nochange)
     struct stat sdata;
     if (last_load_time && (stat(ticket_key_filename, &sdata) >= 0)) {
       if (sdata.st_mtime && sdata.st_mtime <= last_load_time) {
-        Debug("ssl", "ticket key %s has not changed", ticket_key_filename);
+        Debug("ssl_load", "ticket key %s has not changed", ticket_key_filename);
         // No updates since last load
         return true;
       }
@@ -577,12 +688,12 @@ SSLTicketParams::LoadTicket(bool &nochange)
   default_global_keyblock = keyblock;
   load_time               = time(nullptr);
 
-  Debug("ssl", "ticket key reloaded from %s", ticket_key_filename);
+  Debug("ssl_load", "ticket key reloaded from %s", ticket_key_filename);
 #endif
   return true;
 }
 
-void
+bool
 SSLTicketParams::LoadTicketData(char *ticket_data, int ticket_data_len)
 {
   cleanup();
@@ -593,7 +704,12 @@ SSLTicketParams::LoadTicketData(char *ticket_data, int ticket_data_len)
     default_global_keyblock = ssl_create_ticket_keyblock(nullptr);
   }
   load_time = time(nullptr);
+
+  if (default_global_keyblock == nullptr) {
+    return false;
+  }
 #endif
+  return true;
 }
 
 void
@@ -634,7 +750,10 @@ SSLTicketKeyConfig::reconfigure_data(char *ticket_data, int ticket_data_len)
 {
   SSLTicketParams *ticketKey = new SSLTicketParams();
   if (ticketKey) {
-    ticketKey->LoadTicketData(ticket_data, ticket_data_len);
+    if (ticketKey->LoadTicketData(ticket_data, ticket_data_len) == false) {
+      delete ticketKey;
+      return false;
+    }
   }
   configid = configProcessor.set(configid, ticketKey);
   return true;
@@ -647,16 +766,82 @@ SSLTicketParams::cleanup()
   ticket_key_filename = static_cast<char *>(ats_free_null(ticket_key_filename));
 }
 
+void
+cleanup_bio(BIO *&biop)
+{
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wunused-value"
+  BIO_set_close(biop, BIO_NOCLOSE);
+#pragma GCC diagnostic pop
+  BIO_free(biop);
+  biop = nullptr;
+}
+
+void
+SSLConfigParams::updateCTX(const std::string &cert_secret_name) const
+{
+  Debug("ssl_config_updateCTX", "Update cert %s, %p", cert_secret_name.c_str(), this);
+
+  // Instances of SSLConfigParams should access by one thread at a time only.  secret_for_updateCTX is accessed
+  // atomically as a fail-safe.
+  //
+  char const *expected = nullptr;
+  if (!secret_for_updateCTX.compare_exchange_strong(expected, cert_secret_name.c_str())) {
+    if (is_debug_tag_set("ssl_config_updateCTX")) {
+      // As a fail-safe, handle it if secret_for_updateCTX doesn't or no longer points to a null-terminated string.
+      //
+      char const *s{expected};
+      for (; *s && (std::size_t(s - expected) < cert_secret_name.size()); ++s) {
+      }
+      Debug("ssl_config_updateCTX", "Update cert, indirect recusive call caused by call for %.*s", int(s - expected), expected);
+    }
+    return;
+  }
+
+  // Clear the corresponding client CTXs.  They will be lazy loaded later
+  Debug("ssl_load", "Update cert %s", cert_secret_name.c_str());
+  this->clearCTX(cert_secret_name);
+
+  // Update the server cert
+  SSLMultiCertConfigLoader loader(this);
+  loader.update_ssl_ctx(cert_secret_name);
+
+  secret_for_updateCTX = nullptr;
+}
+
+void
+SSLConfigParams::clearCTX(const std::string &client_cert) const
+{
+  ink_mutex_acquire(&ctxMapLock);
+  for (auto ctx_map_iter = top_level_ctx_map.begin(); ctx_map_iter != top_level_ctx_map.end(); ++ctx_map_iter) {
+    auto ctx_iter = ctx_map_iter->second.find(client_cert);
+    if (ctx_iter != ctx_map_iter->second.end()) {
+      ctx_iter->second = nullptr;
+      Debug("ssl_load", "Clear client cert %s %s", ctx_map_iter->first.c_str(), ctx_iter->first.c_str());
+    }
+  }
+  ink_mutex_release(&ctxMapLock);
+}
+
 shared_SSL_CTX
 SSLConfigParams::getCTX(const char *client_cert, const char *key_file, const char *ca_bundle_file, const char *ca_bundle_path) const
+{
+  return this->getCTX(std::string(client_cert ? client_cert : ""), std::string(key_file ? key_file : ""), ca_bundle_file,
+                      ca_bundle_path);
+}
+
+shared_SSL_CTX
+SSLConfigParams::getCTX(const std::string &client_cert, const std::string &key_file, const char *ca_bundle_file,
+                        const char *ca_bundle_path) const
 {
   shared_SSL_CTX client_ctx = nullptr;
   std::string top_level_key, ctx_key;
+  ctx_key = client_cert;
   ts::bwprint(top_level_key, "{}:{}", ca_bundle_file, ca_bundle_path);
-  ts::bwprint(ctx_key, "{}:{}", client_cert, key_file);
 
-  Debug("ssl", "Load client cert %s %s", client_cert, key_file);
+  Debug("ssl_client_ctx", "Look for client cert %s %s", top_level_key.c_str(), ctx_key.c_str());
 
+  ink_mutex_acquire(&ctxMapLock);
   auto ctx_map_iter = top_level_ctx_map.find(top_level_key);
   if (ctx_map_iter != top_level_ctx_map.end()) {
     auto ctx_iter = ctx_map_iter->second.find(ctx_key);
@@ -664,29 +849,85 @@ SSLConfigParams::getCTX(const char *client_cert, const char *key_file, const cha
       client_ctx = ctx_iter->second;
     }
   }
+  ink_mutex_release(&ctxMapLock);
 
+  BIO *biop     = nullptr;
+  X509 *cert    = nullptr;
+  EVP_PKEY *key = nullptr;
   // Create context if doesn't exists
   if (!client_ctx) {
+    Debug("ssl_client_ctx", "Load new cert for %s %s", top_level_key.c_str(), ctx_key.c_str());
     client_ctx = shared_SSL_CTX(SSLInitClientContext(this), SSLReleaseContext);
 
-    if (client_cert) {
-      // Set public and private keys
-      if (!SSL_CTX_use_certificate_chain_file(client_ctx.get(), client_cert)) {
-        SSLError("failed to load client certificate from %s", client_cert);
-        goto fail;
-      }
-      if (!key_file || key_file[0] == '\0') {
-        key_file = client_cert;
-      }
-      if (!SSL_CTX_use_PrivateKey_file(client_ctx.get(), key_file, SSL_FILETYPE_PEM)) {
-        SSLError("failed to load client private key file from %s", key_file);
+    // Set public and private keys
+    if (!client_cert.empty()) {
+      std::string secret_data;
+      std::string secret_key_data;
+
+      // Fetch the client_cert data
+      std::string completeSecretPath{Layout::get()->relative_to(this->clientCertPathOnly, client_cert)};
+      std::string completeKeySecretPath{!key_file.empty() ? Layout::get()->relative_to(this->clientKeyPathOnly, key_file) : ""};
+      secrets.getOrLoadSecret(completeSecretPath, completeKeySecretPath, secret_data, secret_key_data);
+      if (secret_data.empty()) {
+        SSLError("failed to access cert %s", client_cert.c_str());
         goto fail;
       }
 
-      if (!SSL_CTX_check_private_key(client_ctx.get())) {
-        SSLError("client private key (%s) does not match the certificate public key (%s)", key_file, client_cert);
+      biop = BIO_new_mem_buf(secret_data.data(), secret_data.size());
+
+      cert = PEM_read_bio_X509(biop, nullptr, nullptr, nullptr);
+      if (!cert) {
+        SSLError("failed to load cert %s", client_cert.c_str());
         goto fail;
       }
+      if (!SSL_CTX_use_certificate(client_ctx.get(), cert)) {
+        SSLError("failed to attach client certificate from %s", client_cert.c_str());
+        goto fail;
+      }
+      X509_free(cert);
+
+      // Continue to fetch certs to associate intermediate certificates
+      cert = PEM_read_bio_X509(biop, nullptr, nullptr, nullptr);
+      while (cert) {
+        if (!SSL_CTX_add_extra_chain_cert(client_ctx.get(), cert)) {
+          SSLError("failed to attach client chain certificate from %s", client_cert.c_str());
+          goto fail;
+        }
+        // Do not free cert becasue SSL_CTX_add_extra_chain_cert takes ownership of cert if it succeeds, unlike
+        // SSL_CTX_use_certificate.
+        cert = PEM_read_bio_X509(biop, nullptr, nullptr, nullptr);
+      }
+
+      cleanup_bio(biop);
+
+      const std::string &key_file_name = (secret_key_data.empty()) ? client_cert : key_file;
+
+      // If there is a separate key file, fetch the new content
+      // otherwise, continue on with the cert data and hope for the best
+      if (!secret_key_data.empty()) {
+        biop = BIO_new_mem_buf(secret_key_data.data(), secret_key_data.size());
+      } else {
+        biop = BIO_new_mem_buf(secret_data.data(), secret_data.size());
+      }
+
+      key = PEM_read_bio_PrivateKey(biop, nullptr, nullptr, nullptr);
+      if (!key) {
+        SSLError("failed to load client private key file from %s", key_file_name.c_str());
+        goto fail;
+      }
+      if (!SSL_CTX_use_PrivateKey(client_ctx.get(), key)) {
+        SSLError("failed to use client private key file from %s", key_file_name.c_str());
+        goto fail;
+      }
+      EVP_PKEY_free(key);
+      key = nullptr;
+
+      if (!SSL_CTX_check_private_key(client_ctx.get())) {
+        SSLError("client private key (%s) does not match the certificate public key (%s)", key_file_name.c_str(),
+                 client_cert.c_str());
+        goto fail;
+      }
+      cleanup_bio(biop);
     }
 
     // Set CA information for verifying peer cert
@@ -713,6 +954,15 @@ SSLConfigParams::getCTX(const char *client_cert, const char *key_file, const cha
   return client_ctx;
 
 fail:
+  if (biop) {
+    cleanup_bio(biop);
+  }
+  if (cert) {
+    X509_free(cert);
+  }
+  if (key) {
+    EVP_PKEY_free(key);
+  }
   return nullptr;
 }
 
