@@ -21,9 +21,7 @@ Test.Summary = '''
 Test HTTP/2 with httpspec
 '''
 
-Test.SkipUnless(
-    Condition.HasProgram("h2spec", "h2spec need to be installed on system for this test to work"),
-)
+Test.SkipUnless(Condition.HasProgram("h2spec", "h2spec need to be installed on system for this test to work"),)
 Test.ContinueOnFail = True
 
 # ----
@@ -39,20 +37,17 @@ ts = Test.MakeATSProcess("ts", enable_tls=True, enable_cache=False)
 # add ssl materials like key, certificates for the server
 ts.addDefaultSSLFiles()
 
-ts.Disk.remap_config.AddLine(
-    'map / http://127.0.0.1:{0}'.format(httpbin.Variables.Port)
-)
-ts.Disk.ssl_multicert_config.AddLine(
-    'dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key'
-)
-ts.Disk.records_config.update({
-    'proxy.config.http.insert_request_via_str': 1,
-    'proxy.config.http.insert_response_via_str': 1,
-    'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
-    'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
-    'proxy.config.diags.debug.enabled': 0,
-    'proxy.config.diags.debug.tags': 'http',
-})
+ts.Disk.remap_config.AddLine('map / http://127.0.0.1:{0}'.format(httpbin.Variables.Port))
+ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
+ts.Disk.records_config.update(
+    {
+        'proxy.config.http.insert_request_via_str': 1,
+        'proxy.config.http.insert_response_via_str': 1,
+        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
+        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
+        'proxy.config.diags.debug.enabled': 0,
+        'proxy.config.diags.debug.tags': 'http',
+    })
 
 # ----
 # Test Cases
